module AST where

import List
import Data.Generics
import Data.Typeable

import Pretty
import SemanticObject
import Identifiers
import Logging
import qualified MosmlPrelude

data Annotation a = NotAnnotated
                  | Annotated a    
                    deriving (Show,Eq,Typeable,Data)

fromAnnotated (Annotated x) = x
fromAnnotated NotAnnotated = error "fromAnnotated failed"

isAnnotated NotAnnotated = False
isAnnotated _ = True

--
-- Modules language
--

data SigBod = OpaqueTyp TypId
            | ManifestTyp TypId SimTyp
            | ValSpec ValId ValTyp
              deriving (Show,Eq,Typeable,Data)

data SigExp = SigExp [SigBod]
              deriving (Show,Eq,Typeable,Data)

data StrBod = TypDef TypId SimTyp (Annotation SemSimTyp)
            | ValDef ValId ValExp
              deriving (Show,Eq,Typeable,Data)

data PrimStrExp = StrBod [StrBod]
              deriving (Show,Eq,Typeable,Data)

getStrBod :: PrimStrExp -> [StrBod]
getStrBod (StrBod l) = l

data StrExp = StrVar StrId (Annotation SemStr)
            | PrimStrExp PrimStrExp
            | FunctorApp FunId [StrId] (Annotation (Int, SemFun, [SemStr]))
              deriving (Show,Eq,Typeable,Data)

data SealedPrimStrExp = 
    SealedPrim PrimStrExp (Annotation SemStr) SigExp (Annotation SemSig)
  | UnsealedPrim PrimStrExp (Annotation SemStr)
    deriving (Show,Eq,Typeable,Data)

data SealedStrExp = Sealed StrExp (Annotation SemExStr) SigExp (Annotation SemSig)
                  | Unsealed StrExp (Annotation SemExStr)
                    deriving (Show,Eq,Typeable,Data)

data ProgElem = StrDef StrId SealedStrExp
              | FunctorDef FunId [(StrId, SigExp)] SealedPrimStrExp 
                  (Annotation SemFun)
                deriving (Show,Eq,Typeable,Data)

data Prog = Prog [ProgElem]
          -- needed to run the program with Moscow ML. Adds a small
          -- prelude and the invocation
          -- of the main function to the pretty printed output
          | ProgInvokeMain [ProgElem]  
            deriving (Show,Eq,Typeable,Data)

--
-- Core language
--

-- Types
--
data SimTyp = SimTypTyvar SimTyvar          -- 'a
            | SimTypTycon TyconId [SimTyp]  -- T u1 ... uN
            | TmpTyApp SimTyp SimTyp        -- only used during parsing
            | ModTyp TypId                  -- t
            | QModTyp StrId TypId           -- X.t
              deriving (Show,Eq,Typeable,Data)

data ValTyp = ValTyp [SimTyvar] SimTyp
              deriving (Show,Eq,Typeable,Data)

isTyvar (SimTypTyvar _) = True
isTyvar _         = False

freeTyvars :: SimTyp -> [SimTyvar]
freeTyvars (SimTypTyvar a) = [a]
freeTyvars (SimTypTycon _ l) = foldr union [] (map freeTyvars l)
freeTyvars _ = []

-- Expressions
--
data Binop = Add | Sub | Mult | Div | And | Or
           | Eq | Neq | Leq | Lt | Geq | Gt
             deriving (Show,Eq,Typeable,Data)

data ValExp = IntExp Integer              -- 42
            | BinExp Binop ValExp ValExp  -- 1 + 2
            | CoreVarExp CoreId           -- c
            | LamExp CoreId ValExp        -- \c -> e
            | AppExp ValExp ValExp        -- e e
            | LetExp CoreId ValExp ValExp -- let c = e in e
            | IfExp ValExp ValExp ValExp  -- if e then e else e
            | ModVarExp ValId             -- x
            | QModVarExp StrId ValId      -- X.x
              deriving (Show,Eq,Typeable,Data)

isFunctional (LamExp _ _) = True
isFunctional _            = False

--
-- Pretty printing
--

instance Pretty a => Pretty (Annotation a) where
    ppr NotAnnotated  = empty
    ppr (Annotated a) = char '<' <> ppr a <> char '>'

instance Pretty SigBod where
    ppr (OpaqueTyp t)     = text "type" <+> ppr t
    ppr (ManifestTyp t u) = text "type" <+> ppr t <=> ppr u
    ppr (ValSpec x v)     = text "val" <+> ppr x <:> ppr v

instance Pretty SigExp where
    ppr (SigExp [])  = text "sig end"
    ppr (SigExp bod) = text "sig" $$ nest 2 (vcat (map ppr bod)) $$
                       text "end"

freshId = "x__PRIVATE_FUNCTIONAL__"

instance Pretty StrBod where
    ppr (TypDef t u ann) = text "type" <+> ppr t <=> ppr u <> ppr ann
    ppr (ValDef x e) | isFunctional e
        -- Moscow ML is so stupid that it does not classify a function
        -- inside parenthesis as "functional"
        = text "val rec" <+> ppr x <=> text "fn" <+> text freshId <+> text "=>" 
          <+> ppr e <+> text freshId
                     | otherwise
        = text "val" <+> ppr x <=> ppr e

instance Pretty PrimStrExp where
    ppr (StrBod [])            = text "struct end"
    ppr (StrBod bod)           = text "struct" $$ nest 2 (vcat (map ppr bod))
                                 $$ text "end"
 
instance Pretty StrExp where
    ppr (StrVar x ann)         = ppr x <> ppr ann
    ppr (PrimStrExp prim)      = ppr prim
    ppr (FunctorApp f xs ann)  = ppr f <> ppr ann <+> 
                                 hsep (map (\ a -> parens $ ppr a) xs)

instance Pretty SealedPrimStrExp where
    ppr (UnsealedPrim str ann)        = ppr str <> ppr ann
    ppr (SealedPrim str ann sig ann') = ppr str <> ppr ann <+> text ":>" 
                                        <+> ppr sig <> ppr ann'

instance Pretty SealedStrExp where
    ppr (Unsealed str ann)        = ppr str <> ppr ann
    ppr (Sealed str ann sig ann') = ppr str <> ppr ann <+> text ":>" 
                                    <+> ppr sig <> ppr ann'

instance Pretty ProgElem where
    ppr (StrDef x str) = text "structure" <+> ppr x <=> ppr str
    ppr (FunctorDef f args bod ann) = 
        text "functor" <+> ppr f <> ppr ann <+>
        hsep (map (\ (a,b) -> parens $ ppr a <:> ppr b) args) <=> ppr bod

instance Pretty Prog where
    ppr (Prog l) = vcat (punctuate semi $ map ppr l)
    ppr (ProgInvokeMain l) = ppr (Prog l) <> semi $$ text "Main.main;"

instance Pretty SimTyp where
    ppr (SimTypTyvar a)      = ppr a
    ppr (SimTypTycon (TyconId "->") [t1,t2]) 
        = parens (ppr t1 <+> text "->" <+> ppr t2)
    ppr (SimTypTycon (TyconId "->") _) 
        = niceError ("function space type constructor with wrong number of arguments")
    ppr (SimTypTycon n ts)   = parens (vcat (map ppr ts) <+> ppr n)
    ppr (ModTyp t)           = ppr t
    ppr (QModTyp x t)        = ppr x <> char '.' <> ppr t
    ppr (TmpTyApp _ _)       = niceError ("TmpTyApp still present in AST")

instance Pretty ValTyp where
    ppr (ValTyp _ u) = ppr u

instance Pretty Binop where
    ppr Add   = text "+"
    ppr Sub   = text "-"
    ppr Mult  = text "*"
    ppr Div   = text "/"
    ppr And   = text "andalso"
    ppr Or    = text "orelse"
    ppr Eq    = text "=="
    ppr Neq   = text "/="
    ppr Leq   = text "<="
    ppr Lt    = text "<"
    ppr Geq   = text ">="
    ppr Gt    = text ">"
    ppr op    = error ("No pretty printer for binary operator " ++ show op)

instance Pretty ValExp where
    ppr (IntExp i)          = integer i
    ppr (BinExp op e1 e2)   = parens (ppr e1 <+> ppr op <+> ppr e2)
    ppr (CoreVarExp c)      = ppr c
    ppr (LamExp c e)        = parens (text "fn" <+> ppr c <+> text "=>"
                                      <+> ppr e)
    ppr (AppExp e1 e2)      = parens (ppr e1 <+> ppr e2)
    ppr (LetExp c e1 e2)    = parens (text "let val" <+> ppr c <=> ppr e1
                                      <+> text "in" <+> ppr e2 <+> text "end")
    ppr (IfExp e1 e2 e3)    = parens (text "if" <+> ppr e1 <+> text "then" <+>
                                      ppr e2 <+> text "else" <+> ppr e3)
    ppr (ModVarExp x)       = ppr x
    ppr (QModVarExp x y)    = ppr x <> char '.' <> ppr y
                                      

--
-- Utilities
--

stripSealing :: Data a => a -> a
stripSealing x = everywhere (mkT stripSealingFromStrexp `extT` 
                             stripSealingFromPrimStrexp) x
    where stripSealingFromStrexp :: SealedStrExp -> SealedStrExp
          stripSealingFromStrexp (Sealed strexp ann sig _) = Unsealed strexp ann
          stripSealingFromStrexp x = x
          stripSealingFromPrimStrexp :: SealedPrimStrExp -> SealedPrimStrExp
          stripSealingFromPrimStrexp (SealedPrim strexp ann sig _) = 
              UnsealedPrim strexp ann
          stripSealingFromPrimStrexp x = x

removeTmpTyApps :: Data a => a -> a
removeTmpTyApps x = everywhere (mkT flattenTmpTyApps) x
    where flattenTmpTyApps :: SimTyp -> SimTyp
          flattenTmpTyApps t@(TmpTyApp _ _) = flatten t
          flattenTmpTyApps x = x
          flatten (TmpTyApp (SimTypTycon id []) t) = SimTypTycon id [t]
          flatten (TmpTyApp t1 t2) = 
              case flatten t1 of
                SimTypTycon id l -> SimTypTycon id (l++[t2])
                _ -> niceError "error flattening type application"
          flatten u = u --niceError "error flattening type application"
