module Annotation (annotate) where

import Control.Monad.State

import Logging
import AST
import SemanticObject
import Pretty
import Identifiers

--
-- Annotation monad
--

type AN = State ANState

data ANState = ANState 
    { an_functorInvocation :: Int
    , an_semanticProgram :: SemProgPlus -- the "abstract" semantic objects
    }

emptyANState p = ANState { an_functorInvocation = 1
                         , an_semanticProgram = p }

runAN :: SemProgPlus -> AN a -> a
runAN p an = evalState an (emptyANState p)

newFunctorInvocation :: AN Int
newFunctorInvocation = 
    do i <- gets an_functorInvocation
       modify (\s -> s { an_functorInvocation = i+1 })
       return i

lookupSem :: String -> AN SemToplevelPlus
lookupSem id = 
    do (SemProgPlus assoc) <- gets an_semanticProgram
       case lookup id assoc of
         Just sem -> return sem
         Nothing -> niceError ("semantic object for " ++ show id ++ " not found")

--
-- The following two functions return abstract semantic objects
--
lookupSemStr :: StrId -> AN SemStr
lookupSemStr (StrId id) = 
    do sem <- lookupSem id
       case sem of
         SemStructurePlus _ (SemExStr _ x) _ -> return x
         y -> niceError ("found " ++ show y ++ " for " ++ show id ++ 
                         ", but this is not a semantic structure")

lookupSemFun :: FunId -> AN SemFun
lookupSemFun (FunId id) = 
    do sem <- lookupSem id
       case sem of
         SemFunctorPlus _ x _ -> return x
         y -> niceError ("found " ++ show y ++ " for " ++ show id ++ 
                         ", but this is not a semantic functor")


--
-- Annotation code
--

annotateStrBod :: StrBod -> AssocList TypId SemSimTyp -> StrBod
annotateStrBod (TypDef t u _) assoc = 
    case lookup t assoc of
      Just sem -> TypDef t u (Annotated sem)
      Nothing -> niceError ("no semantic object found for type component " 
                            ++ showPpr t)
annotateStrBod valdef _ = valdef

annotatePrimStrExp :: PrimStrExp           -- source structure expression
                   -> (Concrete SemStr)  -- CONCRETE semantic structure!
                   -> AN PrimStrExp        -- annotated structure expression
annotatePrimStrExp (StrBod bod) (Concrete (SemStr typs _)) =
    do let bod' = map (\b -> annotateStrBod b typs) bod
       return (StrBod bod')

annotateStrExp :: StrExp              -- source structure expression
               -> (Concrete SemExStr) -- CONCRETE semantic structure!
               -> AN StrExp           -- annotated structure expression
annotateStrExp (StrVar x _) (Concrete (SemExStr [] sem)) =
    do sem' <- lookupSemStr x
       if sem /= sem'
          then niceError ("sem /= sem'. sem = " ++ showPpr sem ++ ", sem' = " 
                          ++ showPpr sem')
          else return (StrVar x (Annotated sem))
annotateStrExp (PrimStrExp e) (Concrete (SemExStr [] semStr)) = 
    do prim <- annotatePrimStrExp e (Concrete semStr)
       return $ PrimStrExp prim
annotateStrExp (FunctorApp f args _) _ =
    do i <- newFunctorInvocation
       semFun <- lookupSemFun f
       semArgs <- mapM lookupSemStr args
       return (FunctorApp f args (Annotated (i, semFun, semArgs)))

annotateSealedPrimStrExp :: SealedPrimStrExp        -- source structure expression
                         -> SemExStr            -- semantic structure with sealing
                         -> (Concrete SemExStr) -- semantic structure WITHOUT sealing
                         -> AN SealedPrimStrExp     -- annotated structure expression
annotateSealedPrimStrExp (SealedPrim strexp _ sig _) semExStrAbstr 
                         (Concrete (SemExStr [] semStrConcr)) =
    do strexp' <- annotatePrimStrExp strexp (Concrete semStrConcr)
       return (SealedPrim strexp' (Annotated $ semStrConcr) 
                      sig (Annotated (semExStr2semSig semExStrAbstr)))
-- doesn't matter wether we pick the concrete or the abstract semantic structure
-- in the unsealed case
annotateSealedPrimStrExp (UnsealedPrim strexp _) _ 
                         (Concrete (SemExStr [] semStrConcr)) = 
    do strexp' <- annotatePrimStrExp strexp (Concrete semStrConcr)
       return (UnsealedPrim strexp' (Annotated $ semStrConcr))

annotateSealedStrExp :: SealedStrExp        -- source structure expression
                     -> SemExStr            -- semantic structure with sealing
                     -> (Concrete SemExStr) -- semantic structure WITHOUT sealing
                     -> AN SealedStrExp     -- annotated structure expression
annotateSealedStrExp (Sealed strexp _ sig _) semExStrAbstr semExStrConcr  = 
    do strexp' <- annotateStrExp strexp semExStrConcr
       return (Sealed strexp' (Annotated $ fromConcrete semExStrConcr) 
                      sig (Annotated (semExStr2semSig semExStrAbstr)))
-- doesn't matter wether we pick the concrete or the abstract semantic structure
-- in the unsealed case
annotateSealedStrExp (Unsealed strexp _) _ semExStr = 
    do strexp' <- annotateStrExp strexp semExStr
       return (Unsealed strexp' (Annotated $ fromConcrete semExStr))

annotateToplevel :: ProgElem 
                 -> SemToplevelPlus
                 -> AN ProgElem
annotateToplevel (StrDef x strexp) (SemStructurePlus x' semExStr concreteSemExStr)
    | x == x'
    = do strexp' <- annotateSealedStrExp strexp semExStr concreteSemExStr
         return $ StrDef x strexp'
annotateToplevel (FunctorDef f args strexp _) 
                 (SemFunctorPlus f' semFun concreteSemFun) 
    | f == f'
    = do strexp' <- annotateSealedPrimStrExp strexp (getBodySemExStr semFun)
                           (fmap getBodySemExStr concreteSemFun)
         return $ FunctorDef f args strexp' (Annotated semFun)
annotateTopevel x y = niceError ("cannot annotate toplevel construct " ++ showPpr x ++
                                 " with abstract/concrete semantic objects " 
                                 ++ showPpr y)

annotate :: Prog        -- the "source" program
         -> SemProgPlus -- semantic objects
         -> Prog        -- the annotated program
annotate (Prog elems) sp@(SemProgPlus assoc) =
    let semElems = map snd assoc
        anElems = runAN sp $
                  mapM (uncurry annotateToplevel) (zip elems semElems)
        in Prog anElems
