module Main where

import Control.Monad            ( when )
import System.Console.GetOpt
import System.IO                ( hFlush, stdin, stdout, stderr, hPutStrLn,
                                  hGetContents )

import Parser
import System
import Pretty
import Logging
import Mosml
import Annotation
import Translation
import PhracBridge
import Constants

dump' :: [Flag] -> String -> IO ()
dump' fs s =
    if not (null s)
       then do hPutStrLn stdout s
               if ExitAfterDump `elem` fs
                  then exitWith ExitSuccess
                  else return ()
       else return ()

main = 
    do args <- getArgs
       prog <- getProgName
       (flags,fname) <- parseArgs args prog
   
       let on = \x -> elem x flags
       let dump p = dump' flags p
   
       let ll = case logLevel flags of
                  Nothing -> Warn
                  Just l -> l
           in initLogging ll UseStderr
   
       info ("parsing " ++ fname)
       pres <- parseProg fname
       ast <- case pres of
                Right x -> return x
                Left err -> fatal ("Parsing of source code failed!\n\n" ++
                                   show err)
   
       when (on DumpRawParse) $ dump (show ast)
       when (on DumpMosml) $ dump (show.ppr $ ast)
       when (on RunMosml) $ do runProg ast (mosmlCmd flags)
                               exitWith ExitSuccess

       info ("getting semantic objects")
       semProg <- getSemanticObjects ast (mosmlCmd flags)
       when (on DumpSemantic) $ 
            dump ("Semantic objects:\n" ++ (show.ppr) semProg)
       
       info ("translating and running")

       let hs = translate (annotate ast semProg)
       when (on DumpHS) $
            dump ("Haskell code:\n" ++ showPpr hs)

       run (logLevel flags) hs
       
       return ()
       


-- ---------------------------------------------------------------------
-- Handling command line flags

logLevel :: [Flag] -> Maybe LogLevel
logLevel [] = Nothing
logLevel (LogLevel l:_) = Just l
logLevel (_:rest) = logLevel rest

mosmlCmd :: [Flag] -> MosmlCmd
mosmlCmd [] = "mosml"
mosmlCmd (MosmlCmdFlag c : _) = c
mosmlCmd (_:rest) = mosmlCmd rest

data Flag   
    = Version
    | LogLevel LogLevel
    | Help 
    | RunMosml
    | DumpMosml
    | DumpRawParse
    | DumpSemantic
    | ExitAfterDump
    | MosmlCmdFlag String
    | DumpHS
  deriving Eq

--
-- associate concrete flags with their abstract representation
--
opts :: [OptDescr Flag]
opts = 
    [ Option ['v'] ["version"] (NoArg Version) "version string"
    , Option ['?'] ["help"] (NoArg Help) "this message"
    , Option []    ["debug"] (NoArg (LogLevel Debug)) "enable debug messages"
    , Option []    ["verbose"] (NoArg (LogLevel Info)) 
            "be verbose (but not as much as with `--debug')"
    , Option []    ["dump-haskell"] (NoArg DumpHS)
            "dump the generated haskell source code"
    , Option []    ["run-mosml"] (NoArg RunMosml)
            "run the program with Moscow ML, then exit"
    , Option []    ["exit-after-dump"] (NoArg ExitAfterDump) 
            "terminate the program after dumping"
    , Option []    ["dump-mosml"] (NoArg DumpMosml)
            "dump Moscow ML source code"
    , Option []    ["dump-ast"] (NoArg DumpRawParse)
            "dump raw abstract syntax"
    , Option []    ["dump-semantic"] (NoArg DumpSemantic)
            "dump semantic objects"
    , Option []    ["mosml"] (ReqArg MosmlCmdFlag "PATH")
            "path to `mosml', the interactive system of Moscow ML"
    ]

--
-- parse the args.
--
-- we also short circuit here, for the known cases of --version or --help
--
parseArgs :: [String] -> String -> IO ([Flag],FilePath)
parseArgs argv prog = 
    case (getOpt Permute opts argv) of
      (flags,fs,[]) -> case () of { _ 
         | Version `elem` flags -> do putStrLn ml2hsVersion
                                      exitWith ExitSuccess
         | Help `elem` flags -> do putStrLn usage
                                   exitWith ExitSuccess
         | otherwise -> do if length fs == 1
                              then return (flags,fs!!0)
                              else do putStrLn usage
                                      exitFailure
      }
      (_,_,err) -> do putStrLn (concat err ++ usage)
                      exitFailure
    where 
    usage = usageInfo ("Usage: " ++ prog ++ " [OPTION...] file") opts

