module Mosml where

import AST
import Pretty
import Logging
import SemanticObject
import SemanticObjectParser
import Identifiers
import Process
import MosmlPrelude

type MosmlCmd = String

mosmlArgs = ["-liberal", "-P", "full"]

suffix = "__PRIVATE_UNSEALED__"

rewriteProg :: Prog -> Prog
rewriteProg (Prog l) = Prog $ concatMap rewrite' l
    where rewrite' def@(StrDef (StrId x) sealedExp) = 
              -- we have to be careful: sealedExp may be a functor application,
              -- which generates new type variables. we must NOT duplicate
              -- the functor application
              case sealedExp of
                Unsealed strexp _ ->
                  [StrDef (StrId (x ++ suffix)) (Unsealed strexp NotAnnotated), 
                   StrDef (StrId x) (Unsealed (StrVar (StrId (x ++ suffix)) 
                                               NotAnnotated) NotAnnotated)]
                Sealed strexp _ sigexp _ ->
                  [StrDef (StrId (x ++ suffix)) (Unsealed strexp NotAnnotated), 
                   StrDef (StrId x) (Sealed (StrVar (StrId (x ++ suffix)) 
                                             NotAnnotated) 
                                     NotAnnotated sigexp NotAnnotated)]
          rewrite' def@(FunctorDef (FunId f) args sealedExp _) = 
              let unsealedExp = stripSealing sealedExp
                  in [FunctorDef (FunId (f ++ suffix)) args unsealedExp NotAnnotated,
                      def]

rewriteSemProg :: SemProg -> SemProgPlus
rewriteSemProg (SemProg l) = 
    let l' = rewriteSemToplevel (map snd l)
        assoc = addIds l'
        in SemProgPlus assoc
    where addIds (str@(SemStructurePlus (StrId x) _ _) : rest) =
              (x, str) : addIds rest
          addIds (fun@(SemFunctorPlus (FunId f) _ _) : rest) =
              (f, fun) : addIds rest
          addIds [] = []

strIdOk (StrId x) (StrId x') = (x ++ suffix) == x'
funIdOk (FunId x) (FunId x') = (x ++ suffix) == x'

rewriteSemToplevel :: [SemToplevel] -> [SemToplevelPlus]
rewriteSemToplevel (SemStructure x' semExStr' : SemStructure x semExStr : rest)
    | x `strIdOk` x' = SemStructurePlus x semExStr (Concrete semExStr') :
                      rewriteSemToplevel rest
rewriteSemToplevel (SemFunctor f' semFun' : SemFunctor f semFun : rest)
    | f `funIdOk` f' = SemFunctorPlus f semFun (Concrete semFun') :
                      rewriteSemToplevel rest
rewriteSemToplevel [] = []
rewriteSemToplevel l = niceError ("invalid semantic program: " ++ 
                                  (show $ pprList "\n" (map ppr l)))

getSemanticObjects :: Prog -> MosmlCmd -> IO SemProgPlus
getSemanticObjects prog cmd = 
    do debug ("Now running Moscow ML to get the semantic objects")
       out <- runMosml (rewriteProg prog) cmd
       debug ("now parsing Moscow ML's output")
       case parseSem out of
         Right ast -> return $ rewriteSemProg ast
         Left err -> 
              fatal ("Parsing of semantic objects (Moscow ML output) failed\n\n"
                     ++ show err ++ "\n\n\nOutput of Moscow ML:\n\n" ++ number out)

runProg :: Prog -> MosmlCmd -> IO ()
runProg prog cmd = 
    let prog' = case prog of
                  Prog l -> ProgInvokeMain l
                  p -> p
    in do out <- runMosml prog' cmd
          putStrLn out

runMosml :: Prog -> MosmlCmd -> IO String
runMosml prog cmd =
    do debug ("Moscow ML command: " ++ cmd)
       let src = show $ ppr prog
       let fullSrc = MosmlPrelude.src ++ src
       debug ("Pretty printed (full) source code:\n" ++ fullSrc)
       (out, err, pid) <- popen cmd mosmlArgs (Just fullSrc)
       if not (null err)
          then fatal ("Moscow ML failed!!\n\n==> Input:\n" ++ 
                      src ++
                      "\n\n==>Stdout:\n" ++ out ++ "\n\n==>Stderr:\n" ++ err)
          else debug ("Moscow ML finished successfully, output:\n" ++ out)
       return out

-- adds line numbers
number :: String -> String
number s = let l = lines s
               l' = map (\ (i,x) -> align (length l) i ++ ": " ++ x) (zip [1..] l)
               in unlines l'
    where align n i = let s = show i
                          in take (length (show n) - length s) (repeat ' ') ++ s