module SemanticObject where

import List
import Pretty
import Data.Generics
import Data.Typeable

import Identifiers
import Logging

type AssocList a b = [(a,b)]

data Concrete a = Concrete a
                  deriving (Show,Eq,Typeable,Data)

fromConcrete (Concrete x) = x

instance Functor Concrete where
    fmap f (Concrete x) = Concrete (f x)

data SemProg = SemProg (AssocList String SemToplevel)

data SemProgPlus = SemProgPlus (AssocList String SemToplevelPlus)

data SemToplevel = SemStructure StrId SemExStr
                 | SemFunctor FunId SemFun
                   deriving (Show,Eq,Typeable,Data)

-- The (Concrete XXX) fields contain the semantic object obtained
-- by removing the sealing from the rhs of the declaration
data SemToplevelPlus = SemStructurePlus StrId SemExStr (Concrete SemExStr)
                     | SemFunctorPlus FunId SemFun (Concrete SemFun)
                       deriving (Show,Eq,Typeable,Data)

data SemStr = SemStr { sem_types  :: AssocList TypId SemSimTyp
                     , sem_values :: AssocList ValId SemValTyp }
              deriving (Show,Eq,Typeable,Data)

data SemExStr = SemExStr [SemTyvar] SemStr
                deriving (Show,Eq,Typeable,Data)

getSemStr :: SemExStr -> SemStr
getSemStr (SemExStr _ x) = x

quantifiers :: SemExStr -> [SemTyvar]
quantifiers (SemExStr qs _) = qs

data SemSig = SemSig [SemTyvar] SemStr
              deriving (Show,Eq,Typeable,Data)

semExStr2semSig (SemExStr tyvars semStr) = SemSig tyvars semStr

semSig2semExStr (SemSig tyvars semStr) = SemExStr tyvars semStr

data SemFun = SemFun [SemTyvar] [SemStr] SemExStr
              deriving (Show,Eq,Typeable,Data)

getBodySemExStr :: SemFun -> SemExStr
getBodySemExStr (SemFun _ _ ex) = ex

getBodySemStr :: SemFun -> SemStr
getBodySemStr = getSemStr . getBodySemExStr

getArgs :: SemFun -> [SemStr]
getArgs (SemFun _ args _) = args

getQuantifiedVariables :: SemFun -> [SemTyvar]
getQuantifiedVariables (SemFun vs _ _) = vs

data SemValTyp = SemValTyp [SemSimTyvar] SemSimTyp
                 deriving (Show,Eq,Typeable,Data)

data SemSimTyp = SemSimTypSimTyvar SemSimTyvar
               | SemTycon SemTyconId [SemSimTyp]
               | SemSimTypTyvar SemTyvar
               | SemTmpTyApp SemSimTyp SemSimTyp
                 deriving (Show,Eq,Typeable,Data)

freeSemSimTyvars :: SemSimTyp -> [SemSimTyvar]
freeSemSimTyvars (SemSimTypSimTyvar a) = [a]
freeSemSimTyvars (SemTycon _ l) = foldr union [] (map freeSemSimTyvars l)
freeSemSimTyvars _ = []

arrowTycon = SemTyconId "->"
intTycon = SemTyconId "int"

--
-- Pretty printing
--

instance Pretty SemProg where
    ppr (SemProg pairs) = 
        vcat (map (ppr.snd) pairs)

instance Pretty SemProgPlus where
    ppr (SemProgPlus pairs) = 
        vcat (map (ppr.snd) pairs)

instance Pretty SemToplevel where
    ppr (SemStructure name exStr) = text "structure" <+> ppr name <=> ppr exStr
    ppr (SemFunctor name fun) = text "functor" <+> ppr name <=> ppr fun

instance Pretty SemToplevelPlus where
    ppr (SemStructurePlus name exStr (Concrete concrStr)) = 
        text "structure" <+> ppr name <=> ppr exStr <+> text ", concrete:" <+> 
        ppr concrStr
    ppr (SemFunctorPlus name fun (Concrete concrFun)) = 
        text "functor" <+> ppr name <=> ppr fun <+> text ", concrete:" <+> 
        ppr concrFun

instance Pretty SemStr where
    ppr (SemStr typs vals) = 
        braces $ vcat (punctuate comma $
                       map (pprPair' " = ") typs ++ map (pprPair' " = ") vals)

instance Pretty SemExStr where
    ppr (SemExStr tyvars semStr) = 
        text "Exists" <+> hsep (map ppr tyvars) <> text "." <+> ppr semStr

instance Pretty SemSig where
    ppr (SemSig tyvars semStr) = 
        text "Lambda" <+> hsep (map ppr tyvars) <> text "." <+> ppr semStr
    
instance Pretty SemFun where
    ppr (SemFun tyvars args body) =
        text "/\\" <+> hsep (map ppr tyvars) <> text "." <+>
        vcat (punctuate comma $ map ppr args) $$ text "->" <+> ppr body

instance Pretty SemValTyp where
    ppr (SemValTyp vs u) = 
        (if not (null vs)
            then text "Forall" <+> hcat (map ppr vs) <> text ". "
            else empty) <> ppr u

instance Pretty SemSimTyp where
    ppr (SemTmpTyApp t1 t2) = parens (ppr t1 <+> ppr t2)
    ppr (SemSimTypSimTyvar a) = ppr a
    ppr (SemTycon (SemTyconId "->") [t1,t2]) =
        parens (ppr t1 <+> text "->" <+> ppr t2)
    ppr (SemTycon (SemTyconId "->") _) = 
        error ("function space type constructor with wrong number of arguments")
    ppr (SemTycon n ts) = parens (ppr n <+> vcat (map ppr ts))
    ppr (SemSimTypTyvar a) = ppr a



--
-- Utilities
--

removeSemTmpTyApps :: Data a => a -> a
removeSemTmpTyApps x = everywhere (mkT flattenSemTmpTyApps) x
    where flattenSemTmpTyApps :: SemSimTyp -> SemSimTyp
          flattenSemTmpTyApps (SemTmpTyApp t (SemTycon id [])) = 
              SemTycon id (flatten t)
          flattenSemTmpTyApps t@(SemTmpTyApp _ _) = 
              niceError ("error flattening type application: " ++ showPpr t)
          flattenSemTmpTyApps x = x
          flatten (SemTmpTyApp t1 t2) = t2 : flatten t1
          flatten t = [t]
