module SemanticObjectParser where

import Text.ParserCombinators.Parsec
import qualified Text.ParserCombinators.Parsec.Token as T
import Text.ParserCombinators.Parsec.Language
import Text.ParserCombinators.Parsec.Expr
import Maybe

import Identifiers
import SemanticObject
import Logging

data SemState = SemState
    { typeVariables :: [SemTyvar] }

emptySemState = SemState { typeVariables = [] }

type SemParser a = CharParser SemState a


--
-- Lexer
--

lexer :: T.TokenParser SemState
lexer = T.makeTokenParser 
        (emptyDef
         { reservedNames = ["type", "val",
                            "structure", "functor" ]
         , reservedOpNames = []
         , opLetter = oneOf ""
         , opStart = oneOf ""
         , identStart = (letter <|> char '\'')
         , identLetter = (alphaNum <|> char '_' <|> char '/')
         })

whiteSpace = T.whiteSpace lexer
lexeme     = T.lexeme lexer
symbol     = T.symbol lexer
natural    = T.natural lexer
parens     = T.parens lexer
braces     = T.braces lexer
brackets   = T.brackets lexer
semi       = T.semi lexer
identifier = T.identifier lexer
reserved   = T.reserved lexer
commaSep1  = T.commaSep1 lexer
semiSep    = T.semiSep lexer
commaSep   = T.commaSep lexer
colon      = T.colon lexer
dot        = T.dot lexer

--
-- Parser
--

parseTypeIdentifier :: String -> SemParser SemSimTyp
parseTypeIdentifier s@('\'':_) = return $ SemSimTypSimTyvar (SemSimTyvar s)
parseTypeIdentifier s =
    do state <- getState
       if (SemTyvar s) `elem` typeVariables state
          then return $ SemSimTypTyvar (SemTyvar s)
          else return $ SemTycon (SemTyconId s) []

semSimtyp = buildExpressionParser table factor <?> "semantic simple type"
    where
    table = [ [Infix (return SemTmpTyApp <?> "function application") AssocLeft]
            , [Infix ((do symbol "->"
                          return $ \t1 t2 -> SemTycon arrowTycon [t1,t2])
                      <?> "function space type constructor")
                     AssocRight]]
    factor = parens semSimtyp
         <|> do x <- identifier
                parseTypeIdentifier x

tyvar = 
    do char '\''
       x <- identifier
       return ('\'' : x)

prompt = do symbol "-"
            symbol ">"
       
newTypeNames :: SemParser [SemTyvar]
newTypeNames = try (do symbol "New"
                       symbol "type"
                       symbol "names"
                       colon
                       ids <- commaSep identifier
                       return $ map SemTyvar ids)
           <|> return [] 

data SemStrComponent = TypComponent TypId SemSimTyp
                     | ValComponent ValId SemValTyp

mkSemStr :: [SemStrComponent] -> SemStr
mkSemStr [] = SemStr [] []
mkSemStr (TypComponent t u : rest) =
    let SemStr typs vals = mkSemStr rest
        in SemStr ((t,u):typs) vals
mkSemStr (ValComponent x v : rest) =
    let SemStr typs vals = mkSemStr rest
        in SemStr typs ((x,v):vals)

semStrComponent :: SemParser SemStrComponent
semStrComponent = 
    do reserved "type"
       t <- identifier
       symbol "="
       u <- semSimtyp
       return $ TypComponent (TypId t) u
  <|> do reserved "val"
         xs <- many1 identifier  -- explicitly scoped type variables!
         let x = last xs
         colon
         u <- semSimtyp
         let free = freeSemSimTyvars u
         return $ ValComponent (ValId x) (SemValTyp free u)

{-
- > New type names: t, s
  structure X : {type t = t, type s = s, val x : t}

OR

- > structure Z : {type t = int, val x : 'a -> int}
-}
structureSig :: SemParser (String, SemExStr)
structureSig = 
    do prompt
       fresh <- newTypeNames
       updateState (\state -> state { typeVariables = 
                                      fresh ++ typeVariables state })
       reserved "structure"
       x <- identifier
       colon
       components <- braces (commaSep semStrComponent)
       return (x, SemExStr fresh (mkSemStr components))

{-
- > functor F :
  !t s.{type t = t, type s = s}->!t/1.{type t = t/1}
  ->?t/1 s/1.{type t = t/1, type s = s/1}
-}
functorSig :: SemParser (String, SemFun)
functorSig = 
    do prompt
       reserved "functor"
       f <- identifier
       colon
       sigs <- semStr `sepBy` (try $ symbol "->")
       let (argSigs, resSig) = 
               case reverse sigs of
                 (x:xs) -> (reverse xs, x)
                 [] -> niceError "error parsing functor signatur"
       let argSemStrs = map snd argSigs
           argTyvars = concatMap fst argSigs
           resSemStr = snd resSig
           resTyvars = fst resSig
       return (f, SemFun argTyvars argSemStrs (SemExStr resTyvars resSemStr))
    where
    semStr :: SemParser ([SemTyvar], SemStr)
    semStr = 
        do vars <- option [] (do (symbol "!" <|> symbol "?")
                                 res <- many identifier
                                 dot
                                 return (map SemTyvar res))
           updateState (\state -> state { typeVariables = 
                                          vars ++ typeVariables state })
           components <- braces (commaSep semStrComponent)
           return (vars, mkSemStr components)

--
-- Parser for the output of Moscow ML's interactive system
--

moscowMLversionString = "Moscow ML version 2.01 (January 2004)"
parseGreeting :: SemParser ()
parseGreeting = 
    do symbol moscowMLversionString
              <?> ("wrong version of Moscow ML. Expected: " ++
                   moscowMLversionString)
       symbol "Enter `quit();' to quit."
              <?> "Unexpected 2nd output of interactive system"
       symbol "- > val it = () : unit"
              <?> "Is the Prelude loaded??"
       return ()

parseInteractiveOutput :: SemParser SemProg
parseInteractiveOutput = 
    do whiteSpace
       parseGreeting
       res <- many (   try (do (name, sig) <- structureSig
                               return $ Just (name, SemStructure (StrId name) sig))
                   <|> do (name, sig) <- functorSig
                          return $ Just (name, SemFunctor (FunId name) sig))
       eof
       return (SemProg (removeSemTmpTyApps $ catMaybes res))

parseSem :: String -> Either ParseError SemProg
parseSem src = runParser parseInteractiveOutput emptySemState "" src
                 
