module Translation where

import Prelude hiding (lookup)
import qualified Prelude

import qualified Data.Map as Map
import Data.Char
import Control.Monad.State
import List ( intersect )

import qualified Phrac.ParseSyntax as HS
import qualified Phrac.Builtins as HSBuiltins

import AST
import SemanticObject
import Identifiers
import Pretty
import Logging

type Map = Map.Map


--
-- Identifier supply
--

type HSTyconId = HS.TypeId
type HSClassId = HS.ClassId
type HSASynId = HS.AssocTypeId
type HSVarId = HS.ValId
type HSTypVar = HS.TypeVarId

strId2TyconId :: StrId -> HSTyconId
strId2TyconId (StrId x) = "TX_" ++ x

strId2ClassId :: StrId -> HSClassId
strId2ClassId (StrId x) = "CX_" ++ x

strIdTypId2ASynId :: StrId -> TypId -> HSASynId
strIdTypId2ASynId (StrId x) (TypId t) = "SXt_" ++ x ++ "_" ++ t

strIdValId2VarId :: StrId -> ValId -> HSVarId
strIdValId2VarId (StrId x) (ValId y) = "zXy_" ++ x ++ "_" ++ y

funId2TyconId :: FunId -> HSTyconId
funId2TyconId (FunId f) = "TF_" ++ f

funIdInt2TyconId :: FunId -> Int -> HSTyconId
funIdInt2TyconId (FunId f) k = "TF" ++ show k ++ "_" ++ f

funId2ClassId :: FunId -> HSClassId
funId2ClassId (FunId f) = "CF_" ++ f

funId2ClassId_arg :: FunId -> HSClassId
funId2ClassId_arg (FunId f) = "CFarg_" ++ f

funIdTypId2ASynId :: FunId -> TypId -> HSASynId
funIdTypId2ASynId (FunId f) (TypId t) = "SFt_" ++ f ++ "_" ++ t

funIdValId2VarId :: FunId -> ValId -> HSVarId
funIdValId2VarId (FunId f) (ValId x) = "SFx_" ++ f ++ "_" ++ x

funIdIntTypeId2ASynId :: FunId -> Int -> TypId -> HSASynId
funIdIntTypeId2ASynId (FunId f) k (TypId t) = "SF" ++ show k ++ "t_" ++ f ++ "_" ++ t

funIdIntValId2VarId :: FunId -> Int -> ValId -> HSVarId
funIdIntValId2VarId (FunId f) k (ValId x) = "SF" ++ show k ++ "x_" ++ f ++ "_" ++ x

coreId2VarId :: CoreId -> HSVarId
coreId2VarId (CoreId c) = "zc_" ++ c

semSimTyvar2TypVar :: SemSimTyvar -> HSTypVar
semSimTyvar2TypVar (SemSimTyvar a) = "a'a_" ++ a

suffixUnused = "__PRIVATE_UNUSED__"

unusedStrId :: StrId -> StrId
unusedStrId (StrId x) = StrId $ x ++ suffixUnused

unusedFunId :: FunId -> FunId
unusedFunId (FunId f) = FunId $ f ++ suffixUnused

suffixFresh = "__PRIVATE_FRESH__"

freshTypVarId1 :: HS.TypeVarId
freshTypVarId1 = "a" ++ suffixFresh

freshTypVarId2 :: HS.TypeVarId
freshTypVarId2 = "b" ++ suffixFresh

freshVarId :: HS.ValId
freshVarId = "z" ++ suffixFresh

-- some additional identifier supply functions
funId2DataId :: FunId -> HS.DataId
funId2DataId (FunId f) = "D_" ++ f

funIdInt2DataId :: FunId -> Int -> HS.DataId
funIdInt2DataId (FunId f) i = "D_" ++ f ++ show i

strId2DataId :: StrId -> HS.DataId
strId2DataId (StrId x) = "D_" ++ x

-- 
-- Environments
--

class Eq key => Environment env key val | env key -> val where
    insert :: key -> val -> env -> env
    lookup :: key -> env -> val
    dom    :: env -> [key]

extend :: Environment e k v => e -> [(k,v)] -> e
extend e l = 
    if not $ null (dom e `intersect` map fst l)
       then niceError "extend failed, domains are not disjoint"
       else foldr (\ (k,v) e' -> insert k v e') e l

(@@) :: Environment e k v => e -> k -> v
env @@ k = lookup k env

forceLookup k m s = 
    case Map.lookup k m of
      Just x -> x
      Nothing ->
          niceError ("Lookup of " ++ showPpr' k ++ " in " ++ s ++ "failed. " ++
                     show (dumpMap m))

forceLookupList k m s = 
    case Prelude.lookup k m of
      Just x -> x
      Nothing ->
          niceError ("Lookup of " ++ showPpr' k ++ " in " ++ s ++ "failed. " ++
                     show (pprList ", " m))


data AccessPathEnv = AccessPathEnv { ape_semTyvars    :: Map SemTyvar HS.AssocType
                                   , ape_localValues  :: Map ValId HS.Exp
                                   , ape_globalValues :: Map (StrId,ValId) HS.Exp }

emptyAccessPathEnv = AccessPathEnv Map.empty Map.empty Map.empty

instance Pretty AccessPathEnv where
    ppr ape = text "{AccessPathEnvironment. Type variables: " <> 
              dumpMap (ape_semTyvars ape) <>
              text ". Local values: " <> dumpMap (ape_localValues ape) <>
              text ". Global values: " <> dumpMap (ape_globalValues ape)

instance Environment AccessPathEnv SemTyvar HS.AssocType where
    insert k v e = e { ape_semTyvars = Map.insert k v (ape_semTyvars e) }
    lookup k e = forceLookup k (ape_semTyvars e) 
                 "AccessPathEnv, type variable map"
    dom e = Map.keys (ape_semTyvars e)

instance Environment AccessPathEnv ValId HS.Exp where
    insert k v e = e { ape_localValues = Map.insert k v (ape_localValues e) }
    lookup k e = forceLookup k (ape_localValues e)
                 "AccessPathEnv, local value map"
    dom e = Map.keys (ape_localValues e)

instance Environment AccessPathEnv (StrId,ValId) HS.Exp where
    insert k v e = e { ape_globalValues = Map.insert k v (ape_globalValues e) }
    lookup k e = forceLookup k (ape_globalValues e)
                 "AccessPathEnv, global value map"
    dom e = Map.keys (ape_globalValues e)

data CodeEnv = CodeEnv { ce_types  :: Map.Map TypId HS.Type
                       , ce_values :: Map.Map ValId HS.Exp }

emptyCodeEnv = CodeEnv Map.empty Map.empty

instance Environment CodeEnv TypId HS.Type where
    insert k v e = e { ce_types = Map.insert k v (ce_types e) }
    lookup k e = forceLookup k (ce_types e)
                 "CodeEnv, type map"
    dom e = Map.keys (ce_types e)

instance Environment CodeEnv ValId HS.Exp where
    insert k v e = e { ce_values = Map.insert k v (ce_values e) }
    lookup k e = forceLookup k (ce_values e)
                 "CodeEnv, value map"
    dom e = Map.keys (ce_values e)

instance Environment SemStr TypId SemSimTyp where
    insert k v e = niceError "lookup not supported from semantic structure"
    lookup k e = forceLookupList k (sem_types e) "Semantic structure, types"
    dom e = map fst (sem_types e)

instance Environment SemExStr TypId SemSimTyp where
    insert k v (SemExStr vars semStr) = SemExStr vars $ insert k v semStr
    lookup k (SemExStr _ semStr) = lookup k semStr
    dom (SemExStr _ semStr) = dom semStr

instance Environment SemStr ValId SemValTyp where
    insert k v e = niceError "lookup not supported from semantic structure"
    lookup k e = forceLookupList k (sem_values e) "Semantic structure, values"
    dom e = map fst (sem_values e)

instance Environment SemExStr ValId SemValTyp where
    insert k v (SemExStr vars semStr) = SemExStr vars $ insert k v semStr
    lookup k (SemExStr _ semStr) = lookup k semStr
    dom (SemExStr _ semStr) = dom semStr

--
-- The translation monad
--
data TMState = TMState { tm_freshVarCounter :: Int }

emptyTMState = TMState { tm_freshVarCounter = 0 }

type TM = StateT TMState IO

runTM :: TM a -> IO a
runTM tm = evalStateT tm emptyTMState

--
-- Program translation
--

translate :: Prog -> HS.Program
translate prog = translateProg prog emptyAccessPathEnv

translateProg :: Prog -> AccessPathEnv -> HS.Program
translateProg (Prog []) phi = HS.Program [] [] [] [mainBind]
    where mainBind = HS.ValBind "main" Nothing [] mainExp toplevelInfo
          mainExp = HS.App (HS.Var "zXy_Main_main") (HS.Con "D_Main" [])

translateProg (Prog (StrDef strId (Unsealed strExp (Annotated semExStr)) : restProg)) 
              phi = 
    let (ddecs, clss, insts) = translateStructureHelper strExp semExStr strId phi
        HS.Program ddecs' clss' insts' vals = translateProg (Prog restProg) phi'
        in HS.Program (ddecs ++ ddecs') (clss ++ clss') (insts ++ insts') vals
    where phi' = phi `extend`
                 [ (alpha, HS.AssocType (assocId $ pick semExStr alpha) [typ])
                    | alpha <- quantifiers semExStr ] `extend`
                 [ ( (strId,y), HS.App (HS.Var (methodId y)) dat )
                   | y <- dom semExStr ]
          typ = HS.TyConstruct (strId2TyconId strId)
          dat = HS.Con (strId2DataId strId) []
          assocId = strIdTypId2ASynId strId
          methodId = strIdValId2VarId strId

translateProg (Prog (StrDef strId (Sealed strExp (Annotated semExStr')
                                          sigExp (Annotated (SemSig p semStr)))
                     : restProg))
              phi =
    let ddec = mkDdec tyconId datconId
        cls = HS.ClassDec classId [a] [{- no constraints-}]
                [HS.AssocTypeSig (assocId t) [a] Nothing | t <- dom semStr]
                [(methodId y, HS.TypeScheme (b y) (HS.QualifiedType [] $
                                                   HS.TyVar a `tyFun` tau y))
                 | y <- dom semStr ]
                toplevelInfo
        inst = HS.ClassInst classId [instType] [{- no constraints-}]
                 -- associated type synonyms
                 ([HS.AssocTypeBind (assocId t) [instType] []
                     (HS.TyAssoc (HS.AssocType (assocIdStar t) [instTypeStar]))
                     HS.AssocTypeNotAbstract
                   | t <- dom semStr, notInP t] ++
                  [HS.AssocTypeBind (assocId t) [instType] []
                     (HS.TyAssoc (HS.AssocType (assocIdStar t) [instTypeStar]))
                     HS.AssocTypeAbstract
                   | t <- dom semStr, inPandPicked t] ++
                  [HS.AssocTypeBind (assocId t) [instType] []
                     (HS.TyAssoc (HS.AssocType (assocIdOfPicked t) [instType]))
                     HS.AssocTypeNotAbstract
                   | t <- dom semStr, inPandNotPicked t])
                 -- methods
                 [HS.ValBind (methodId y) Nothing []
                    (HS.Lam [HS.PWildcard] (HS.Var (methodIdStar y) `HS.App`
                                           (HS.Con datconIdStar [])))
                    toplevelInfo
                  | y <- dom semStr]
                 toplevelInfo
    in HS.Program (ddecs ++ ddec:ddecs') (clss ++ cls:clss') (insts ++ inst:insts') 
                  vals
    where
    (ddecs, clss, insts) = translateStructureHelper strExp semExStr' strIdStar phi
    HS.Program ddecs' clss' insts' vals = translateProg (Prog restProg) phi'
    phi' = phi `extend` [ (alpha, (HS.AssocType (assocId $ pick semStr alpha) 
                                   [instType]))
                          | alpha <- p]
               `extend` [ ((strId, y), HS.App (HS.Var $ methodId y) 
                                              (HS.Con datconId []))
                          | y <- dom semStr ]
    b y = case translateSemValTyp (lookup y semStr) phi'' of
            HS.TypeScheme quants _ -> quants
    tau y = case translateSemValTyp (lookup y semStr) phi'' of
              HS.TypeScheme _ (HS.QualifiedType _ tau) -> tau
    phi'' = phi `extend` [ (alpha, (HS.AssocType (assocId $ pick semStr alpha) 
                                    [HS.TyVar a]))
                           | alpha <- p]
    a = freshTypVarId1
    --
    notInP t = case semStr@@t of
                 SemSimTypTyvar alpha -> not $ alpha `elem` p
                 _ -> True
    inPandPicked t = case semStr@@t of
                       SemSimTypTyvar alpha -> 
                           if not $ alpha `elem` p
                              then False
                              else t == pick semStr alpha
                       _ -> False
    inPandNotPicked t = not (notInP t) && not (inPandPicked t)
    assocIdOfPicked t = 
        case semStr@@t of
          SemSimTypTyvar alpha -> assocId $ pick semStr alpha
          _ -> niceError ("assocIdOfPicked: not a type variable")
    --
    strIdStar = unusedStrId strId
    assocId = strIdTypId2ASynId strId
    assocIdStar = strIdTypId2ASynId strIdStar
    methodId = strIdValId2VarId strId
    methodIdStar = strIdValId2VarId strIdStar
    classId = strId2ClassId strId 
    tyconId = strId2TyconId strId
    tyconIdStar = strId2TyconId strIdStar
    datconId = strId2DataId strId
    datconIdStar = strId2DataId strIdStar
    instType = HS.TyConstruct tyconId
    instTypeStar = HS.TyConstruct tyconIdStar


    
    
translateProg (Prog (FunctorDef funId args (UnsealedPrim ps (Annotated semStr))
                     (Annotated semFun) : restProg)) phi =
    let (ddecs, clss, insts) = translateFunctorHelper ps semStr funId semFun 
                                 (map fst args) funId phi
        HS.Program ddecs' clss' insts' vals = translateProg (Prog restProg) phi      
        in HS.Program (ddecs ++ ddecs') (clss ++ clss') (insts ++ insts') vals

translateProg (Prog (FunctorDef funId args (SealedPrim strExp 
                                            (Annotated semStr')
                                            sigExp 
                                            (Annotated (SemSig p semStr)))
                     (Annotated (SemFun q argSemStr _)) : restProg)) phi = 
    let ddec = mkDdec tyconId datconId
        cls = HS.ClassDec classId [b, a] 
                [HS.ClassConstraint classIdArg [HS.TyVar a]]
                [HS.AssocTypeSig (assocId t) [b,a] Nothing | t <- dom semStr]
                [(methodId y, 
                  HS.TypeScheme (quants y) (HS.QualifiedType [] $
                                            HS.TyVar b `tyFun` 
                                            (HS.TyVar a `tyFun` tau y)))
                 | y <- dom semStr ]
                toplevelInfo
        inst = HS.ClassInst classId [instType, tyvarA] 
                 [HS.CC $ HS.ClassConstraint classIdArg [tyvarA]]
                 -- associated type synonyms
                 ([HS.AssocTypeBind (assocId t) [instType, tyvarA] []
                     (HS.TyAssoc (HS.AssocType (assocIdStar t) 
                                  [instTypeStar, tyvarA]))
                     HS.AssocTypeNotAbstract
                   | t <- dom semStr, notInP t] ++
                  [HS.AssocTypeBind (assocId t) [instType, tyvarA] []
                     (HS.TyAssoc (HS.AssocType (assocIdStar t) 
                                  [instTypeStar, tyvarA]))
                     HS.AssocTypeAbstract
                   | t <- dom semStr, inPandPicked t] ++
                  [HS.AssocTypeBind (assocId t) [instType, tyvarA] []
                     (HS.TyAssoc (HS.AssocType (assocIdOfPicked t) 
                                  [instType, tyvarA]))
                     HS.AssocTypeNotAbstract
                   | t <- dom semStr, inPandNotPicked t])
                 -- methods
                 [HS.ValBind (methodId y) Nothing []
                    (HS.Lam [HS.PWildcard, HS.PVar z] 
                     ((HS.Var (methodIdStar y) `HS.App` (HS.Con datconIdStar []))
                      `HS.App` (HS.Var z)))
                    toplevelInfo
                  | y <- dom semStr]
                 toplevelInfo
    in HS.Program (ddecs ++ ddec:ddecs') (clss ++ cls:clss') (insts ++ inst:insts') 
                  vals
    where
    (ddecs, clss, insts) = translateFunctorHelper strExp semStr' funIdStar
                             (SemFun q argSemStr (SemExStr [] semStr'))
                             (map fst args) funId phi
    HS.Program ddecs' clss' insts' vals = translateProg (Prog restProg) phi
    quants y = case translateSemValTyp (lookup y semStr) phi' of
                 HS.TypeScheme quants _ -> quants
    tau y = case translateSemValTyp (lookup y semStr) phi' of
              HS.TypeScheme _ (HS.QualifiedType _ tau) -> tau
    phi' = phi `extend` [ (alpha, (HS.AssocType (assocIdArg $ pick argSemStr alpha) 
                                   [HS.TyVar a]))
                          | alpha <- q]
               `extend` [ (alpha, (HS.AssocType (assocId $ pick semStr alpha) 
                                   [HS.TyVar b, HS.TyVar a]))
                          | alpha <- p]
    a = freshTypVarId1
    b = freshTypVarId2
    z = freshVarId
    tyvarA = HS.TyVar a
    --
    notInP t = case semStr@@t of
                 SemSimTypTyvar alpha -> not $ alpha `elem` p
                 _ -> True
    inPandPicked t = case semStr@@t of
                       SemSimTypTyvar alpha -> 
                           if not $ alpha `elem` p
                              then False
                              else t == pick semStr alpha
                       _ -> False
    inPandNotPicked t = not (notInP t) && not (inPandPicked t)
    assocIdOfPicked t = 
        case semStr@@t of
          SemSimTypTyvar alpha -> assocId $ pick semStr alpha
          _ -> niceError ("assocIdOfPicked: not a type variable")
    --
    funIdStar = unusedFunId funId
    assocId t = funIdTypId2ASynId funId t
    assocIdArg (i, t) = funIdIntTypeId2ASynId funId i t
    methodId x = funIdValId2VarId funId x
    assocIdStar t = funIdTypId2ASynId funIdStar t
    methodIdStar x = funIdValId2VarId funIdStar x
    classId = funId2ClassId funId
    classIdArg = funId2ClassId_arg funId
    tyconId = funId2TyconId funId
    datconId = funId2DataId funId
    instType = HS.TyConstruct tyconId
    instTypeStar = HS.TyConstruct tyconIdStar
    tyconIdStar = funId2TyconId funIdStar
    datconIdStar = funId2DataId funIdStar

-- helper function

translateFunctorHelper :: PrimStrExp -> SemStr -> FunId -> SemFun
                       -> [StrId] -> FunId -> AccessPathEnv
                       -> ([HS.TypeDef], [HS.ClassDec], [HS.ClassInst])
translateFunctorHelper ps _ f (SemFun q semStrArgs (SemExStr [] semStrRes)) 
                       argIds f' phi = 
    let n = length (semStrArgs) - 1
        ddec = mkDdec tyconId datconId
        clsArg = 
            HS.ClassDec classIdArg [a] [{-no constraints-}]
                [ HS.AssocTypeSig (assocId' (i, t)) [a] Nothing
                  | i <- [0..n], t <- dom (semStrArgs!!i) ]
                [ (methodId' i x, HS.TypeScheme (quants' x i) 
                                    (HS.QualifiedType [] $
                                     HS.TyVar a `tyFun` tau' x i))
                  | i <- [0..n], x <- dom (semStrArgs!!i) ]
                toplevelInfo
        cls = HS.ClassDec classId [b,a] [HS.ClassConstraint classIdArg [HS.TyVar a]]
                [ HS.AssocTypeSig (assocId t) [b, a] Nothing
                  | t <- dom semStrRes ]
                [ (methodId y, HS.TypeScheme (quants y) 
                                 (HS.QualifiedType [] $
                                  HS.TyVar b `tyFun` (HS.TyVar a `tyFun` tau y)))
                  | y <- dom semStrRes ]
                toplevelInfo                
        inst = HS.ClassInst classId [instType, HS.TyVar a] 
                 [HS.CC $ HS.ClassConstraint classIdArg [HS.TyVar a]]
                 [ HS.AssocTypeBind (assocId t) [instType, HS.TyVar a] [] (omega @@ t)
                     HS.AssocTypeNotAbstract
                   | t <- dom semStrRes ]
                 [ HS.ValBind (methodId y) Nothing [] 
                     (HS.Lam [HS.PWildcard, HS.PVar z] (omega @@ y)) toplevelInfo
                    | y <- dom semStrRes ]
                 toplevelInfo
        in ([ddec], [clsArg, cls], [inst])
    where
    quants' x i  = case translateSemValTyp (lookup x (semStrArgs!!i)) phi' of
                     HS.TypeScheme quants _ -> quants
    tau' x i = case translateSemValTyp (lookup x (semStrArgs!!i)) phi' of
                 HS.TypeScheme _ (HS.QualifiedType _ tau) -> tau
    quants y = case translateSemValTyp (lookup y semStrRes) phi' of
                 HS.TypeScheme quants _ -> quants
    tau y = case translateSemValTyp (lookup y semStrRes) phi' of
              HS.TypeScheme _ (HS.QualifiedType _ tau) -> tau
    phi' = phi `extend` [ (alpha, (HS.AssocType (assocId' $ pick semStrArgs alpha) 
                                   [HS.TyVar a]))
                          | alpha <- q]
    omega = translateStrexpBody (getStrBod ps) phi''
    phi'' = phi' `extend` [ ( (argIds!!i, y), HS.Var (methodId' i y) 
                                              `HS.App` HS.Var z )
                            | i <- [0 .. (length semStrArgs - 1)], 
                              y <- dom (semStrArgs!!i) ]
                 `extend` [ ( x, (HS.Var (methodId x) `HS.App` (HS.Con datconId []))
                                 `HS.App` HS.Var z )
                            | x <- dom semStrRes ]
    a = freshTypVarId1
    b = freshTypVarId2
    z = freshVarId
    --                              
    assocId' (i, t) = funIdIntTypeId2ASynId f' i t
    methodId' i y = funIdIntValId2VarId f' i y
    assocId t = funIdTypId2ASynId f t
    methodId x = funIdValId2VarId f x
    classId = funId2ClassId f
    classIdArg = funId2ClassId_arg f'
    tyconId = funId2TyconId f
    datconId = funId2DataId f
    instType = HS.TyConstruct tyconId
   

translateStructureHelper :: StrExp -> SemExStr -> StrId -> AccessPathEnv
                         -> ([HS.TypeDef], [HS.ClassDec], [HS.ClassInst])
translateStructureHelper strExp (SemExStr p semStr) strId phi = 
    let ddec = mkDdec tyconId datconId
        cls = HS.ClassDec classId [a] [{-no constraints-}]
                [ HS.AssocTypeSig (assocId t) [a] Nothing
                  | t <- dom semStr ]
                [ (methodId y, HS.TypeScheme (b y) (HS.QualifiedType [] $
                                                    HS.TyVar a `tyFun` tau y))
                  | y <- dom semStr ]
                toplevelInfo
        inst = HS.ClassInst classId [instType] [{-no constraints-}]
                 [ HS.AssocTypeBind (assocId t) [instType] [] (omega @@ t)
                     HS.AssocTypeNotAbstract
                   | t <- dom semStr ]
                 [ HS.ValBind (methodId y) Nothing [] 
                     (HS.Lam [HS.PWildcard] (omega @@ y)) toplevelInfo
                    | y <- dom semStr ]
                 toplevelInfo
        in (ddecs ++ [ddec], [cls], insts ++ [inst])
    where
    b y = case translateSemValTyp (lookup y semStr) phi' of
            HS.TypeScheme quants _ -> quants
    tau y = case translateSemValTyp (lookup y semStr) phi' of
              HS.TypeScheme _ (HS.QualifiedType _ tau) -> tau
    phi' = phi `extend` [ (alpha, (HS.AssocType (assocId $ pick semStr alpha) 
                                   [HS.TyVar a]))
                          | alpha <- p]
    (omega, ddecs, insts) = translateStrexp strExp phi''
    phi'' = phi `extend` [ (y, HS.App (HS.Var $ methodId y) (HS.Con datconId [])) 
                           | y <- dom semStr ]
    a = freshTypVarId1
    --                              
    assocId = strIdTypId2ASynId strId
    methodId = strIdValId2VarId strId
    classId = strId2ClassId strId 
    tyconId = strId2TyconId strId
    datconId = strId2DataId strId
    instType = HS.TyConstruct tyconId

--
-- Structure expression translation
--

translateStrexp :: StrExp -> AccessPathEnv -> (CodeEnv, [HS.TypeDef], [HS.ClassInst])
translateStrexp (PrimStrExp (StrBod bod)) phi = 
    (translateStrexpBody bod phi, [], [])
translateStrexp (StrVar x (Annotated semStr)) phi =
    let types = sem_types semStr
        values = sem_values semStr
        typIds = map fst types
        valIds = map fst values
        typTrans = map (\t -> translateSemSimTyp t phi) (map snd types)
        valTrans = map (\y -> lookup (x,y) phi) valIds
       in (CodeEnv (Map.fromList (zip typIds typTrans))
                   (Map.fromList (zip valIds valTrans)), 
           [], [])
translateStrexp (FunctorApp id argIds (Annotated (k, semFun, argSems))) phi = 
    let assocBinds = concatMap mkAssocs (zip3 [0..] argSems (getArgs semFun))
        methodBinds = concatMap mkMethods (zip3 [0..] argIds (getArgs semFun))
        ddec = mkDdec argId argDatcon
        inst = HS.ClassInst classId [instType] [{-no constraints-}]
                 assocBinds methodBinds toplevelInfo
        body = getBodySemStr semFun
        omegaTyps = map mkOmegaTyp (map fst (sem_types body))
        omegaVals = map mkOmegaVal (map fst (sem_values body))
       in (CodeEnv (Map.fromList omegaTyps) (Map.fromList omegaVals), [ddec], [inst])
    where
    mkAssocs :: (Int, SemStr{-actual argument-}, SemStr{-formal argument-}) 
             -> [HS.AssocTypeBind]
    mkAssocs (i, sem, sem') = 
        let mk t = let assocId = funIdIntTypeId2ASynId id i t
                       rhs = translateSemSimTyp (lookup t sem) phi
                       in HS.AssocTypeBind assocId [instType] [] rhs
                            HS.AssocTypeNotAbstract
            in map mk (dom sem')
    mkMethods :: (Int, StrId, SemStr) -> [HS.ValBind]
    mkMethods (i, x, sem') =
        let mk y = let methodId = funIdIntValId2VarId id i y
                       rhs = HS.Lam [HS.PWildcard] (lookup (x,y) phi)
                       in HS.ValBind methodId Nothing [] rhs toplevelInfo
            in map mk (dom sem')
    mkOmegaTyp :: TypId -> (TypId, HS.Type)
    mkOmegaTyp t = 
        let assocId = funIdTypId2ASynId id t
            assocTyp = HS.TyAssoc (HS.AssocType assocId [functorType, instType])
            in (t, assocTyp)
    mkOmegaVal :: ValId -> (ValId, HS.Exp)
    mkOmegaVal x = 
        let methodId = funIdValId2VarId id x
            exp = HS.App (HS.App (HS.Var methodId) (HS.Con functorDatcon []))
                         (HS.Con argDatcon [])
            in (x, exp)
    --
    argId = funIdInt2TyconId id k
    argDatcon = funIdInt2DataId id k
    classId = funId2ClassId_arg id 
    instType = HS.TyConstruct argId
    functorType = HS.TyConstruct (funId2TyconId id)
    functorDatcon = funId2DataId id

translateStrexpBody :: [StrBod] -> AccessPathEnv -> CodeEnv
translateStrexpBody elems phi = 
    foldl translateElem emptyCodeEnv elems
    where translateElem :: CodeEnv -> StrBod -> CodeEnv
          translateElem ce (TypDef t _ (Annotated u)) = 
              insert t (translateSemSimTyp u phi) ce
          translateElem ce (ValDef x e) = 
              insert x (translateExp e phi) ce
                                 
          

--
-- Type translation
--

translateSemSimTyp :: SemSimTyp -> AccessPathEnv -> HS.Type
translateSemSimTyp (SemSimTypSimTyvar simTyvar) _ = 
    HS.TyVar (semSimTyvar2TypVar simTyvar)
translateSemSimTyp (SemTycon tyconId args) phi =
    foldl HS.TyApp (hsTycon tyconId) (map (\t -> translateSemSimTyp t phi) args)
translateSemSimTyp (SemSimTypTyvar tyvar) phi = 
    case Map.lookup tyvar (ape_semTyvars phi) of
      Just at -> HS.TyAssoc at
      Nothing -> niceError ("type variable " ++ showPpr' tyvar ++ 
                            " not found in access path environment " ++ showPpr phi)
translateSemSimTyp t@(SemTmpTyApp _ _) _ = 
    niceError ("temporary type application found during translation: "
               ++ showPpr t)

translateSemValTyp :: SemValTyp -> AccessPathEnv -> HS.TypeScheme
translateSemValTyp (SemValTyp tyvars semSimTyp) phi = 
    let quants = map semSimTyvar2TypVar tyvars
        t = translateSemSimTyp semSimTyp phi
        in HS.TypeScheme quants (HS.QualifiedType [] t)


--
-- Expression translation
--

translateExp :: ValExp -> AccessPathEnv -> HS.Exp
translateExp (IntExp i) _ = HS.Const (HS.Number i)
translateExp (BinExp And e1 e2) phi = 
    HS.App (HS.App andHS (translateExp e1 phi)) (translateExp e2 phi)
translateExp (BinExp Or e1 e2) phi = 
    HS.App (HS.App orHS (translateExp e1 phi)) (translateExp e2 phi)
translateExp (BinExp op e1 e2) phi = 
    HS.Prim (translateOp op) [(translateExp e1 phi), (translateExp e2 phi)]
translateExp (CoreVarExp c) _ = 
    HS.Var $ coreId2VarId c
translateExp (LamExp c e) phi =
    HS.Lam [HS.PVar $ coreId2VarId c] (translateExp e phi)
translateExp (AppExp e1 e2) phi = 
    HS.App (translateExp e1 phi) (translateExp e2 phi)
translateExp (LetExp c e1 e2) phi = 
    HS.Letrec [HS.ValBind (coreId2VarId c) Nothing [] (translateExp e1 phi)
               nestedInfo] (translateExp e2 phi)
translateExp (IfExp e1 e2 e3) phi = 
    HS.If (translateExp e1 phi) (translateExp e2 phi) (translateExp e3 phi)
translateExp (ModVarExp x) phi = 
    case Map.lookup x (ape_localValues phi) of
      Just e -> e
      Nothing -> niceError ("No expression found for local structure value " ++
                            showPpr' x ++ " in access path environment " ++ 
                            showPpr phi)
translateExp (QModVarExp x y) phi =
    case Map.lookup (x,y) (ape_globalValues phi) of
      Just e -> e
      Nothing -> niceError ("No expression found for global structure value `" ++
                            show (ppr x <.> ppr y) ++ 
                            "' in access path environment " ++ showPpr phi)

translateOp op = 
    case op of
      Add -> HSBuiltins.IntAddOp
      Sub -> HSBuiltins.IntSubOp
      Mult -> HSBuiltins.IntMulOp
      Div -> HSBuiltins.IntQuotOp
      Eq -> HSBuiltins.IntEqOp
      Neq -> HSBuiltins.IntNeOp
      Leq -> HSBuiltins.IntLeOp
      Lt -> HSBuiltins.IntLtOp
      Geq -> HSBuiltins.IntGeOp
      Gt -> HSBuiltins.IntGtOp
    

--
-- Utilities
--

andHS = HS.Var "&&"
orHS = HS.Var "||"

nestedInfo :: HS.Info
nestedInfo = HS.Info False False

toplevelInfo :: HS.Info
toplevelInfo = HS.Info False True

hsTycon :: SemTyconId -> HS.Type
hsTycon (SemTyconId []) = niceError "empty type constructor identifier"
hsTycon (SemTyconId (c:cs)) = HS.TyConstruct (toUpper c : cs)

mkDdec id datconId = HS.TypeDef id [] [HS.DataConstructor datconId []] toplevelInfo

tyFun :: HS.Type -> HS.Type -> HS.Type
tyFun t1 t2 = ((HS.TyConstruct "->") `HS.TyApp` t1) `HS.TyApp` t2

tyFuns :: [HS.Type] -> HS.Type -> HS.Type
tyFuns [] t = t
tyFuns (t:ts) t2 = tyFun t (tyFuns ts t2)

class Pickable arg res where
    pick :: arg -> SemTyvar -> res

instance Pickable SemStr (Maybe TypId) where
    pick semStr alpha = pick' (sem_types semStr)
        where pick' [] = Nothing
              pick' ((t, SemSimTypTyvar beta) : _) | alpha == beta = Just t
              pick' (_ : rest) = pick' rest 
    
instance Pickable SemStr TypId where
    pick semStr alpha = 
        case pick semStr alpha of
          Nothing ->
              niceError ("pick failed: semantic structure " ++ showPpr semStr ++
                         " does not provide " ++ showPpr' alpha)
          Just t -> t

instance Pickable SemExStr TypId where
    pick (SemExStr _ semStr) alpha = pick semStr alpha

instance Pickable [SemStr] (Int, TypId) where
    pick semStrs alpha = pick' semStrs 0
        where pick' [] _ =
                  niceError ("pick failed: semantic structure " ++ 
                             (show $ pprList ", " semStrs) ++
                             " does not provide " ++ showPpr' alpha)
              pick' (semStr : rest) i = 
                  case pick semStr alpha of
                    Nothing -> pick' rest (i+1)
                    Just t -> (i,t)

    