module AnnotationTest where

import TestFramework
import AST
import Parser
import qualified Mosml
import SemanticObject
import Annotation
import Pretty
import Identifiers

progSrc = unlines 
          [
            "structure X = struct type t = int end :> sig type t end"
          , "functor F (Y: sig type t end) = struct type t = X.t end"
          , "structure Z = F(X)"
          , "functor G (Y: sig type s end) = struct type s = Y.s end \
            \                                :> sig type s end"
          , "structure Z2 = F(X)"
          ]

getAST :: IO Prog
getAST = 
    do case parseString progSrc of
         Right ast -> return ast
         Left err -> assertFailure ("Parsing failed: " ++ show err)

getSemanticObjects :: Prog -> IO SemProgPlus
getSemanticObjects p = 
    do Mosml.getSemanticObjects p "mosml"

semX = SemStr [(TypId "t", SemSimTypTyvar (SemTyvar "t"))] []
semXConcr = SemStr [(TypId "t", SemTycon intTycon [])] []

semF = SemFun [SemTyvar "t/1"] 
              [SemStr [(TypId "t", SemSimTypTyvar (SemTyvar "t/1"))] []] 
              (SemExStr [] semX)

semG = SemFun [SemTyvar "s"] 
              [SemStr [(TypId "s", SemSimTypTyvar (SemTyvar "s"))] []] 
              (SemExStr [SemTyvar "s/1"] $
               SemStr [(TypId "s", SemSimTypTyvar (SemTyvar "s/1"))] [])

semGBodyConcr = SemStr [(TypId "s", SemSimTypTyvar (SemTyvar "s"))] []

checkStructureX (StrDef (StrId "X") (Sealed 
                                     (PrimStrExp 
                                      (StrBod [TypDef (TypId "t") _ annTyp]))
                                     annStr
                                     (SigExp [OpaqueTyp (TypId "t")])
                                     annSig)) =
    do assertEqual (Annotated $ SemTycon intTycon []) annTyp
       assertEqual (Annotated $ SemExStr [] semXConcr) annStr
       assertEqual (Annotated $ SemSig [SemTyvar "t"] semX) annSig
checkStructureX x = 
    do assertFailure (showPpr x)

checkFunctorF (FunctorDef (FunId "F") _ (UnsealedPrim 
                                         (StrBod [TypDef (TypId "t") _ annTyp])
                                         annStr)
               annFun) =
    do assertEqual (Annotated semX) annStr
       assertEqual (Annotated semF) annFun
       assertEqual (Annotated $ SemSimTypTyvar (SemTyvar "t")) annTyp
checkFunctorF f = 
    do assertFailure (showPpr f)

checkStructureZ (StrDef (StrId "Z") 
                 (Unsealed (FunctorApp (FunId "F") [StrId "X"] annApp) annStr)) =
    do assertEqual (Annotated (1, semF, [semX])) annApp
       assertEqual (Annotated (SemExStr [] semX)) annStr
checkStructureZ z =
    do assertFailure (showPpr z)

checkFunctorG (FunctorDef (FunId "G") _ 
               (SealedPrim
                (StrBod [TypDef (TypId "s") _ annTyp]) 
                annStr
                (SigExp [OpaqueTyp (TypId "s")])
                annSig)
               annFun) =
    do assertEqual (Annotated $ SemSimTypTyvar (SemTyvar "s")) annTyp
       assertEqual (Annotated semG) annFun
       assert (isAnnotated annSig)
       assertEqual (getBodySemExStr (fromAnnotated annFun)) 
                   (semSig2semExStr (fromAnnotated annSig))
       assertEqual (Annotated $ semGBodyConcr) annStr

checkFunctorG g = 
    do assertFailure (showPpr g)

checkStructureZ2 (StrDef (StrId "Z2") 
                  (Unsealed (FunctorApp (FunId "F") [StrId "X"] annApp) annStr)) =
    do assertEqual (Annotated (2, semF, [semX])) annApp
       assertEqual (Annotated (SemExStr [] semX)) annStr
checkStructureZ2 z2 = 
    do assertFailure (showPpr z2)

$(tests "annotationTests" [d|

 test1 = 
     do ast <- getAST
        sems <- getSemanticObjects ast
        let Prog progElems = annotate ast sems
        assertEqual 5 $ length (progElems)
        checkStructureX (progElems!!0)
        checkFunctorF (progElems!!1)
        checkStructureZ (progElems!!2)
        checkFunctorG (progElems!!3)
        checkStructureZ2 (progElems!!4)
        return ()

 |])
          